import abc
import json
import sqlite3
import uuid


class DataObject:
    def __init__(self, id: str | None = None):
        self._id = id if id else str(uuid.uuid4())
        self._data = {}

    def __getattr__(self, name):
        return self._data[name]

    def __setattr__(self, name, value):
        if name.startswith("_"):
            super().__setattr__(name, value)
        elif name not in self._data:
            self._data[name] = value
        else:
            raise AttributeError(f"DataObject attribute {name} already exists")


class Beaker(abc.ABC):
    def __init__(self, name: str, model: type, recipe: "Recipe"):
        self.name = name
        self.model = model
        self.recipe = recipe

    def __repr__(self):
        return f"Beaker({self.name}, {self.model.__name__})"

    @abc.abstractmethod
    def items(self):
        pass

    @abc.abstractmethod
    def __len__(self):
        pass

    @abc.abstractmethod
    def add_item(self, item: "T", id: str | None = None) -> None:
        pass

    @abc.abstractmethod
    def reset(self):
        pass

    def add_items(self, items: list["T"]) -> None:
        for item in items:
            self.add_item(item)


class TempBeaker(Beaker):
    def __init__(self, name: str, model: type, recipe: "Recipe"):
        super().__init__(name, model, recipe)
        self._items = []

    def __len__(self):
        return len(self._items)

    def add_item(self, item: "T", id=None) -> None:
        if id is None:
            id = str(uuid.uuid1())
        self._items.append((id, item))

    def items(self):
        yield from self._items

    def reset(self):
        self._items = []


class SqliteBeaker(Beaker):
    def __init__(self, name: str, model: type, recipe: "Recipe"):
        super().__init__(name, model, recipe)
        # create table if it doesn't exist
        self.cursor = self.recipe.db.cursor()
        self.cursor.row_factory = sqlite3.Row
        self.cursor.execute(
            f"CREATE TABLE IF NOT EXISTS {self.name} (uuid TEXT PRIMARY KEY, data JSON)"
        )

    def items(self):
        self.cursor.execute(f"SELECT uuid, data FROM {self.name}")
        data = self.cursor.fetchall()
        for item in data:
            yield item["uuid"], self.model(**json.loads(item["data"]))

    def __len__(self):
        self.cursor.execute(f"SELECT COUNT(*) FROM {self.name}")
        return self.cursor.fetchone()[0]

    def add_item(self, item: "T", id: str | None = None) -> None:
        if id is None:
            id = str(uuid.uuid1())
            print("UUID", id, item)
        self.cursor.execute(
            f"INSERT INTO {self.name} (uuid, data) VALUES (?, ?)",
            (id, item.model_dump_json()),
        )
        self.recipe.db.commit()

    def reset(self):
        self.cursor.execute(f"DELETE FROM {self.name}")
        self.recipe.db.commit()
